import copy
import os
import time

import numpy as np
from PyFlyt.core import Aviary
from scipy.optimize import linear_sum_assignment


class Simulator:
    """
    This is a virtual version of the swarm_controller or drone_controller code.
    Intended to be used to visualize how the real drones will fly.

    Control is done using linear velocity setpoints and yawrate:
        vx, vy, vz, vr
    States is full linear position and yaw
        x, y, z, r
    """

    def __init__(self, start_pos, start_orn):
        # we use a custom drone that is accurate to the real model
        drone_options = dict()
        drone_options["model_dir"] = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "../models/"
        )
        drone_options["drone_model"] = "cf2x"

        # instantiate the digital twin
        self.env = Aviary(
            drone_type="quadx",
            start_pos=start_pos,
            start_orn=start_orn,
            render=True,
            drone_options=drone_options,
        )
        self.set_pos_control(True)
        self.env.set_armed([0] * self.env.num_drones)

        # keep track of runtime
        self.steps = 0

    def reshuffle(self, new_pos, new_orn):
        """
        reshuffle the drones given a new start_pos such that all drones map to the new start_pos cleanly
        """
        # if start pos is given, reassign to get drones to their positions automatically
        assert (
            new_pos.shape == new_orn.shape
        ), "start_pos must have same shape as start_orn"
        assert (
            len(new_pos) == self.num_drones
        ), "must have same number of drones as number of drones"
        assert (
            new_pos[0].shape[0] == 3
        ), "start pos must have only xyz, start orn must have only pqr"

        # compute cost matrix
        cost = abs(
            np.expand_dims(self.states[:, :3], axis=0) - np.expand_dims(new_pos, axis=1)
        )
        cost = np.sum(cost, axis=-1)

        # compute optimal assignment using Hungarian algo
        _, reassignment = linear_sum_assignment(cost)
        self.env.drones = [self.env.drones[i] for i in reassignment]

        # send setpoints
        setpoints = np.concatenate(
            (new_pos, np.expand_dims(new_orn[:, -1], axis=-1)), axis=-1
        )
        self.set_pos_control(True)
        self.set_setpoints(setpoints)

        cost = np.choose(reassignment, cost.T)
        return cost

    def set_setpoints(self, setpoints: np.ndarray):
        """
        setpoints is a num_drones x 4 array, where the 4 corresponds to x, y, z, r or vx, vy, vz, vr
        """
        # the setpoints in the digital twin has the last two dims flipped
        temp = copy.deepcopy(setpoints[:, -2])
        setpoints[:, -2] = copy.deepcopy(setpoints[:, -1])
        setpoints[:, -1] = temp
        self.env.set_all_setpoints(setpoints)

    def step(self):
        self.steps += 1
        self.env.step()

    def set_pos_control(self, setting):
        """sets entire swarm to fly using pos control"""
        self.env.set_mode(7 if setting else 6)

    def get_states(self):
        raw_states = np.array(self.env.all_states)
        states = np.zeros((self.num_drones, 4))
        states[:, :-1] = copy.deepcopy(raw_states[:, -1, :])
        states[:, -1] = copy.deepcopy(raw_states[:, 1, -1])

        return states

    def sleep(self, seconds: float):
        for _ in range(int(seconds / self.env.update_period)):
            self.step()

    def arm(self, settings):
        self.env.set_armed(settings)

    def end(self):
        self.arm([0] * self.num_drones)
        time.sleep(3)
        exit()

    @property
    def states(self):
        return self.get_states()

    @property
    def num_drones(self):
        return self.env.num_drones

    @property
    def elapsed_time(self):
        return self.env.physics_period * self.steps



import time
from typing import List

import numpy as np
from scipy.optimize import linear_sum_assignment

from .drone_controller import DroneController


class SwarmController:
    """
    Class for controlling a swarm of Crazyflie UAVs.
    Realistically, only about 5 drones per dongle supported.
        Required: URIs
        - correspond to list of URIs for the drones.
        Optional: start_pos and start_orn
        - correspond to desired initial positions of drones
            don't have to worry about drone numbering,
            assignment is solved using Hungarian algorithm.
            Only the last value of start_orn (yaw) is used.
            Position control automatically set if this is supplied.
    """

    def __init__(self, URIs: List[str]):
        self.UAVs = [DroneController(URI, in_swarm=True) for URI in URIs]
        time.sleep(1)
        print(f"Swarm with {self.num_drones} drones ready to go...")
        time.sleep(1)

    def reshuffle(self, new_pos, new_orn):
        """
        reshuffle the drones given a new start_pos such that all drones map to the new start_pos cleanly
        """
        assert (
            new_pos.shape == new_orn.shape
        ), "start_pos must have same shape as start_orn"
        assert (
            len(new_pos) == self.num_drones
        ), "must have same number of drones as number of drones"
        assert (
            new_pos.shape[1] == 3
        ), "start pos must have only xyz, start orn must have only pqr"

        # compute cost matrix
        cost = abs(
            np.expand_dims(self.states[:, :3], axis=0) - np.expand_dims(new_pos, axis=1)
        )
        cost = np.sum(cost, axis=-1)

        # compute optimal assignment using Hungarian algo
        _, reassignment = linear_sum_assignment(cost)
        self.UAVs = [self.UAVs[i] for i in reassignment]

        # send setpoints
        setpoints = np.concatenate(
            (new_pos, np.expand_dims(new_orn[:, -1], axis=-1)), axis=-1
        )
        self.set_pos_control(True)
        self.set_setpoints(setpoints)

        cost = np.choose(reassignment, cost.T)
        return cost

    @property
    def num_drones(self):
        """number of drones in swarm"""
        return len(self.UAVs)

    @property
    def position_estimate(self):
        """list of position estimate for each drone"""
        return np.stack([UAV.position_estimate for UAV in self.UAVs], axis=0)

    @property
    def states(self):
        return self.position_estimate

    def set_pos_control(self, setting):
        """sets entire swarm to fly using pos control"""
        for UAV in self.UAVs:
            UAV.set_pos_control(setting)

    def arm(self, masks):
        """arms and starts control loop for drones with masks set to 1"""
        assert len(masks) == len(
            self.UAVs
        ), "masks length must be equal to number of drones"

        for mask, UAV in zip(masks, self.UAVs):
            if mask:
                UAV.start()
            else:
                UAV.stop()

    def end(self):
        """disarms each drone and closes all connections"""
        for UAV in self.UAVs:
            UAV.end()
        time.sleep(1)

    def set_setpoints(self, setpoints: np.ndarray):
        """sets setpoints for each drone, setpoints must be ndarray where len(setpoints) == len(UAVs)"""
        assert len(setpoints) == len(
            self.UAVs
        ), "number of setpoints must be equal to number of drones"

        for setpoint, UAV in zip(setpoints, self.UAVs):
            UAV.set_setpoint(setpoint)

    def sleep(self, seconds: float):
        time.sleep(seconds)



"""Setup CrazyFlyt."""
import os

from setuptools import setup


def package_files(directory):
    """package_files.

    Args:
        directory: directory of non-python files
    """
    paths = []
    for path, directories, filenames in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


setup(
    name="CrazyFlyt",
    package_data={"CrazyFlyt": package_files("CrazyFlyt/models/")},
)



import math
import threading
import time

import cflib.crtp
import numpy as np
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.positioning.motion_commander import MotionCommander
from cflib.utils import uri_helper


class DroneController:
    """controller for a single drone"""

    def __init__(self, URI, in_swarm=False):
        self.period = 1 / 40.0
        URI = uri_helper.uri_from_env(default=URI)

        self.running = False
        self.flow_deck_attached = False

        self.position_estimate = np.array([0.0, 0.0, 0.0, 0.0])
        self.setpoint = np.array([0.0, 0.0, 0.0, 0.0])

        self.pos_control = False
        self.rad_to_deg = np.array([1.0, 1.0, 1.0, math.pi / 180.0])

        # make connection
        self.scf = None
        try:
            cflib.crtp.init_drivers()
            self.scf = SyncCrazyflie(URI, cf=Crazyflie(rw_cache="./cache"))
            self.scf.open_link()
        except Exception as e:
            print(f"Failed to open link with Flier on {URI}, {e}.")

        # update the onboard PIDs
        self.param_set("posCtlPid", "xKp", 1.2)
        self.param_set("posCtlPid", "yKp", 1.2)
        self.param_set("posCtlPid", "zKp", 1.0)
        self.param_set("posCtlPid", "zKi", 0.2)

        # logging thread
        self.logging_thread = LogConfig(name="Position", period_in_ms=10)
        self.logging_thread.add_variable("stateEstimate.x", "float")
        self.logging_thread.add_variable("stateEstimate.y", "float")
        self.logging_thread.add_variable("stateEstimate.z", "float")
        self.logging_thread.add_variable("stateEstimate.yaw", "float")
        self.scf.cf.log.add_config(self.logging_thread)
        self.logging_thread.data_received_cb.add_callback(self._log_callback)

        # start the logging thread automatically
        self.logging_thread.start()

        # start drone control automatically
        self.control_thread = threading.Thread(name="background", target=self._control)
        self.control_thread.setDaemon(True)
        self.control_thread.start()

        # delay a bit to let things stabilize if not in swarm
        print(f"Flier on {URI} ready to rock and roll...")
        if not in_swarm:
            time.sleep(3)

    def start(self):
        """starts the control loop"""
        self.running = True

    def stop(self):
        """stops the control loop"""
        self.running = False

    def end(self):
        """stops control loop and ends connection with drone"""
        self.running = False
        self.logging_thread.stop()
        self.scf.close_link()

    def set_pos_control(self, setting):
        """set True if pos control is desired, otherwise vel control is used"""
        self.pos_control = setting

    def set_setpoint(self, setpoint):
        """sets the setpoint for flight"""
        self.setpoint = setpoint

    def sleep(self, seconds: float):
        time.sleep(seconds)

    def _control(self):
        """control loop, NOT to be called in main"""
        while True:
            if self.running:
                if self.pos_control:
                    self.scf.cf.commander.send_position_setpoint(
                        *(self.setpoint * self.rad_to_deg)
                    )
                else:
                    self.scf.cf.commander.send_velocity_world_setpoint(
                        *(self.setpoint * self.rad_to_deg)
                    )
            else:
                self.scf.cf.commander.send_stop_setpoint()

            time.sleep(self.period)

    def _log_callback(self, timestamp, data, logconf):
        """logging callback, NOT to be called in main"""
        self.position_estimate[0] = data["stateEstimate.x"]
        self.position_estimate[1] = data["stateEstimate.y"]
        self.position_estimate[2] = data["stateEstimate.z"]
        self.position_estimate[3] = data["stateEstimate.yaw"] / 180.0 * math.pi

    def _update_param_callback(self, name, value):
        pass
        # print('The crazyflie has parameter ' + name + ' set to: ' + value)

    def param_set(self, groupstr, namestr, value):
        full_name = groupstr + "." + namestr

        self.scf.cf.param.add_update_callback(
            group=groupstr, name=namestr, cb=self._update_param_callback
        )
        time.sleep(1)
        self.scf.cf.param.set_value(full_name, value)
